from django import forms
from django.contrib.auth import get_user_model
from django.contrib.auth.forms import UserCreationForm
from django.core.exceptions import ValidationError

from ddm.models import ResearchProfile, DonationProject, DonationBlueprint, DonationInstruction
from ddm.auth import email_is_valid


User = get_user_model()


class ProjectCreateForm(forms.ModelForm):
    secret = forms.CharField(min_length=10, max_length=150, required=False)

    class Meta:
        model = DonationProject
        fields = ['name', 'slug', 'super_secret', 'owner']
        widgets = {'owner': forms.HiddenInput()}

    field_order = ['name', 'slug', 'super_secret', 'secret']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'initial' in kwargs:
            self.fields['owner'].initial = kwargs['initial']['owner']

    def clean(self):
        super_secret = self.data.get('super_secret', False)
        secret = self.data.get('secret', None)
        if super_secret and secret in ['', None]:
            raise ValidationError('Super secret project needs a secret.')
        super().clean()

    def save(self, commit=True):
        project = super().save(commit=False)
        if project.super_secret:
            project.secret_key = self.data['secret']
        project.save()
        return project


class DdmUserCreationForm(UserCreationForm):
    email = forms.EmailField(required=True)

    class Meta:
        model = User
        fields = ('username', 'first_name', 'last_name', 'email', 'password1', 'password2')

    def clean_email(self):
        email = self.cleaned_data['email']
        if not email_is_valid(email):
            raise forms.ValidationError(
                'Only researchers with a valid UZH e-mail address can register.'
            )
        return email

    def save(self, commit=True):
        user = super().save(commit=False)
        user.email = self.cleaned_data['email']
        if commit:
            user.save()
            ResearchProfile.objects.create(user=user)
        return user


class ResearchProfileConfirmationForm(forms.ModelForm):
    confirmed = forms.BooleanField()

    class Meta:
        model = ResearchProfile
        fields = ['user']
        widgets = {'user': forms.HiddenInput()}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.expected_user_id = kwargs['initial']['user']
        self.fields['user'].initial = self.expected_user_id
        self.fields['confirmed'].initial = True
        self.fields['confirmed'].widget = forms.HiddenInput()

    def is_valid(self):
        if super().is_valid():
            return self.cleaned_data['confirmed']
        else:
            return False

    def save(self):
        """
        Ensure that user.id is not changed manually (e.g., by altering the HTML).
        """
        if self.cleaned_data['user'].pk != self.expected_user_id:
            raise ValidationError('User is not set as expected.', code='not allowed')
        return super().save()


class BlueprintEditForm(forms.ModelForm):

    class Meta:
        model = DonationBlueprint
        fields = ['name', 'exp_file_format', 'expected_fields', 'extracted_fields',
                  'zip_blueprint', 'regex_path']
        widgets = {
            'expected_fields': forms.Textarea(attrs={'rows': 3}),
            'extracted_fields': forms.Textarea(attrs={'rows': 3}),
            'regex_path': forms.Textarea(attrs={'rows': 3}),
        }

    def clean(self):
        zip_relation = self.data.get('zip_blueprint', False)
        regex = self.data.get('regex_path', None)
        if zip_relation and regex in ['', None]:
            raise ValidationError(
                'Donation Blueprints that belong to a zip blueprint must define '
                'a regex pattern.'
            )
        super().clean()



from django.test import TestCase
from django.urls import reverse
from django.contrib.auth import get_user_model
from ddm.models import ResearchProfile, DonationProject
from ddm.auth import email_is_valid, user_is_owner, user_is_permitted
from ddm.tests.base import TestData


User = get_user_model()


class TestAuthenticationFlow(TestData, TestCase):

    def test_email_is_valid(self):
        valid_email = 'abc@mail.com'
        invalid_email = 'abc@liam.com'
        self.assertTrue(email_is_valid(valid_email))
        self.assertFalse(email_is_valid(invalid_email))

    def test_user_is_permitted_without_permission(self):
        self.assertFalse(user_is_permitted(self.users['no_permission']['user']))

    def test_user_is_permitted_with_ignore_email_restriction_true(self):
        self.users['no_permission']['profile'].ignore_email_restriction = True
        self.users['no_permission']['profile'].save()
        self.assertTrue(user_is_permitted(self.users['no_permission']['user']))
        self.users['no_permission']['profile'].ignore_email_restriction = False
        self.users['no_permission']['profile'].save()

    def test_user_is_owner(self):
        dp = DonationProject.objects.create(name='test-project', slug='test',
                                            owner=self.users['base']['profile'])
        self.assertTrue(user_is_owner(self.users['base']['user'], dp.pk))
        self.assertFalse(user_is_owner(self.users['no_profile']['user'], dp.pk))

    def test_login_redirect_without_profile(self):
        response = self.client.post(
            reverse('ddm-login'),
            data=self.users['no_profile']['credentials'],
            follow=True
        )
        self.assertRedirects(response, reverse('ddm-register'))

    def test_login_redirect_with_profile(self):
        response = self.client.post(
            reverse('ddm-login'),
            data=self.users['base']['credentials'],
            follow=True
        )
        self.assertRedirects(response, reverse('project-list'))

    def test_register_redirect_with_profile(self):
        self.client.login(**self.users['base']['credentials'])
        response = self.client.get(reverse('ddm-register'), follow=True)
        self.assertRedirects(response, reverse('project-list'))

    def test_register_redirect_after_registration_form_valid(self):
        self.client.login(**self.users['no_profile']['credentials'])
        response = self.client.post(
            reverse('ddm-register'),
            data={'confirmed': True, 'user': self.users['no_profile']['user'].pk},
            follow=True
        )
        self.assertRedirects(response, reverse('project-list'))

    def test_register_redirect_after_registration_form_invalid(self):
        self.client.login(**self.users['no_profile']['credentials'])
        response = self.client.post(
            reverse('ddm-register'),
            data={'confirmed': False, 'user': self.users['no_profile']['user'].pk},
            follow=True
        )
        self.assertRedirects(response, reverse('ddm-no-permission'))

    def test_register_view_creates_research_profile(self):
        self.assertFalse(ResearchProfile.objects.filter(user=self.users['no_profile']['user']).exists())
        self.client.login(**self.users['no_profile']['credentials'])
        self.client.post(
            reverse('ddm-register'),
            data={'confirmed': True, 'user': self.users['no_profile']['user'].pk},
            follow=True
        )
        self.assertTrue(ResearchProfile.objects.filter(user=self.users['no_profile']['user']).exists())

    def test_ddm_no_permission_view_200(self):
        self.client.login(**self.users['no_permission']['credentials'])
        response = self.client.get(
            reverse('ddm-no-permission'), follow=True)
        self.assertEqual(response.status_code, 200)

    def test_ddm_no_permission_view_redirect_to_login(self):
        response = self.client.get(
            reverse('ddm-no-permission'), follow=True)
        self.assertRedirects(response, reverse('ddm-login'))

    def test_ddm_no_permission_view_redirect_to_project_list(self):
        self.client.login(**self.users['base']['credentials'])
        response = self.client.get(
            reverse('ddm-no-permission'), follow=True)
        self.assertRedirects(response, reverse('project-list'))

    def test_create_user_view_redirect_authenticated_with_profile(self):
        self.client.login(**self.users['base']['credentials'])
        response = self.client.get(
            reverse('ddm-create-user'), follow=True)
        self.assertRedirects(response, reverse('project-list'))

    def test_create_user_view_redirect_authenticated_without_profile(self):
        self.client.login(**self.users['no_profile']['credentials'])
        response = self.client.get(
            reverse('ddm-create-user'), follow=True)
        self.assertRedirects(response, reverse('ddm-register'))

    def test_create_user_view_creates_user_and_profile(self):
        self.assertFalse(User.objects.filter(email='new@mail.com').exists())
        self.assertFalse(ResearchProfile.objects.filter(user__email='new@mail.com').exists())
        response = self.client.post(
            reverse('ddm-create-user'),
            data={'username': 'new_user', 'email': 'new@mail.com',
                  'password1': 'Z-6THLAqmd7H5gyUs.8dZt', 'password2': 'Z-6THLAqmd7H5gyUs.8dZt'},
            follow=True
        )
        self.assertTrue(User.objects.filter(email='new@mail.com').exists())
        self.assertTrue(ResearchProfile.objects.filter(user__email='new@mail.com').exists())
        self.assertRedirects(response, reverse('ddm-login'))

    def test_logout_view_redirect_not_logged_in(self):
        response = self.client.get(reverse('ddm-logout'), follow=True)
        self.assertRedirects(response, reverse('ddm-login'))



from django.contrib.auth import get_user_model
from django.contrib.auth.views import PasswordChangeView, PasswordChangeDoneView
from django.contrib.messages.views import SuccessMessageMixin
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.views.generic import DetailView

from ddm.auth import user_is_permitted
from ddm.models import ResearchProfile

User = get_user_model()


class ProfileDetailView(SuccessMessageMixin, DetailView):
    template_name = 'ddm/project_admin/profile/profile_detail.html'
    model = User
    success_message = 'Profile updated.'

    def dispatch(self, request, *args, **kwargs):
        if not request.user.is_authenticated:
            return redirect('ddm-login')
        elif not user_is_permitted(request.user):
            return redirect('ddm-no-permission')
        elif not ResearchProfile.objects.filter(user=request.user).exists():
            return redirect('ddm-register')
        return super().dispatch(request, *args, **kwargs)

    def get_object(self, queryset=None):
        return self.request.user

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['profile'] = ResearchProfile.objects.filter(user=self.object).first()
        return context


class ProfileChangePasswordView(SuccessMessageMixin, PasswordChangeView):
    template_name = 'ddm/project_admin/generic/page_with_form.html'
    success_url = reverse_lazy('ddm-pw-changed')
    success_message = 'Password changed.'


class ProfilePasswordChangedView(PasswordChangeDoneView):
    template_name = 'ddm/project_admin/profile/password_changed.html'



from django.contrib.auth import get_user_model
from django.test import TestCase, override_settings
from django.utils import timezone

from ddm.models import (
    ResearchProfile, DonationProject, DonationBlueprint, ZippedBlueprint,
    DonationInstruction, MatrixQuestion, SingleChoiceQuestion, OpenQuestion, Participant
)


User = get_user_model()


@override_settings(DDM_SETTINGS={'EMAIL_PERMISSION_CHECK':  r'.*(\.|@)mail\.com$', })
class TestData(TestCase):
    @classmethod
    def setUpTestData(cls):
        # Users
        cls.users = {}

        super_user_credentials = {'username': 'super_user', 'password': '123', 'email': 'super@mail.com'}
        super_user = User.objects.create_superuser(**super_user_credentials)
        super_user_profile = ResearchProfile.objects.create(user=super_user)
        cls.users['super'] = {
            'credentials': super_user_credentials,
            'user': super_user,
            'profile': super_user_profile}

        base_user_credentials = {'username': 'base_user', 'password': '123', 'email': 'base@mail.com'}
        base_user = User.objects.create_user(**base_user_credentials)
        base_user_profile = ResearchProfile.objects.create(user=base_user)
        cls.users['base'] = {
            'credentials': base_user_credentials,
            'user': base_user,
            'profile': base_user_profile}

        base_user2_credentials = {'username': 'base_user2', 'password': '123', 'email': 'base2@mail.com'}
        base_user2 = User.objects.create_user(**base_user2_credentials)
        base_user2_profile = ResearchProfile.objects.create(user=base_user2)
        cls.users['base2'] = {
            'credentials': base_user2_credentials,
            'user': base_user2,
            'profile': base_user2_profile}

        base_user3_credentials = {'username': 'base_user3', 'password': '123', 'email': 'base3@mail.com'}
        base_user3 = User.objects.create_user(**base_user3_credentials)
        base_user3_profile = ResearchProfile.objects.create(user=base_user3)
        cls.users['base3'] = {
            'credentials': base_user3_credentials,
            'user': base_user3,
            'profile': base_user3_profile}

        user_wo_profile_credentials = {'username': 'no_prof', 'password': '123', 'email': 'noprof@mail.com'}
        user_wo_profile = User.objects.create_user(**user_wo_profile_credentials)
        cls.users['no_profile'] = {
            'credentials': user_wo_profile_credentials,
            'user': user_wo_profile,
            'profile': None}

        user_wo_permission_credentials = {'username': 'no_per', 'password': '123', 'email': 'noperm@liam.com'}
        user_wo_permission = User.objects.create_user(**user_wo_permission_credentials)
        user_wo_permission_profile = ResearchProfile.objects.create(user=user_wo_permission)
        cls.users['no_permission'] = {
            'credentials': user_wo_permission_credentials,
            'user': user_wo_permission,
            'profile': user_wo_permission_profile}

        # Projects
        cls.project_base = DonationProject.objects.create(
            name='Base Project', slug='base', owner=cls.users['base']['profile'])
        cls.project_base2 = DonationProject.objects.create(
            name='Base2 Project', slug='base-2', owner=cls.users['base2']['profile'])
        cls.project_no_perm = DonationProject.objects.create(
            name='No Perm Project', slug='no-perm', owner=cls.users['no_permission']['profile'])

        # Blueprints
        cls.don_bp = DonationBlueprint.objects.create(
            project=cls.project_base,
            name='donation blueprint',
            expected_fields='"a", "b"',
            extracted_fields='"a"'
        )
        cls.zip_bp = ZippedBlueprint.objects.create(
            name='zipped blueprint',
            project=cls.project_base
        )

        # Questions
        cls.open_quest = OpenQuestion.objects.create(
            project=cls.project_base,
            blueprint=cls.don_bp,
            name='open question',
            variable_name='open_question'
        )
        cls.sc_quest = SingleChoiceQuestion.objects.create(
            project=cls.project_base,
            blueprint=cls.don_bp,
            name='sc question',
            variable_name='sc_question'
        )
        cls.matrix_quest = MatrixQuestion.objects.create(
            project=cls.project_base,
            blueprint=cls.don_bp,
            name='matrix question',
            variable_name='matrix_question'
        )
        cls.instruction = DonationInstruction.objects.create(
            text='some text',
            index=1,
            blueprint=cls.don_bp
        )

        # Participant
        cls.participant_base = Participant.objects.create(
            project=cls.project_base,
            start_time=timezone.now()
        )



from django.contrib.auth import get_user_model
from django.test import TestCase

from ddm.forms import ProjectCreateForm
from ddm.models import ResearchProfile


User = get_user_model()


class TestProjectCreateForm(TestCase):
    @classmethod
    def setUpTestData(cls):
        user_credentials = {'username': 'user', 'password': '123', 'email': 'base@mail.com'}
        user = User.objects.create_user(**user_credentials)
        cls.user_profile = ResearchProfile.objects.create(user=user)

    def test_form_creates_normal_project(self):
        post_data = {'name': 'normal project', 'slug': 'normal-project',
                     'super_secret': False, 'secret': 'supersecretsecret',
                     'owner': self.user_profile}
        form = ProjectCreateForm(post_data)
        project = form.save()
        self.assertEqual(b'', project.public_key)

    def test_form_creates_super_secret_project(self):
        post_data = {'name': 'secret-project', 'slug': 'secret-project',
                     'super_secret': True, 'secret': 'supersecretsecret',
                     'owner': self.user_profile}
        form = ProjectCreateForm(post_data)
        project = form.save()
        self.assertNotEqual(b'', project.public_key)

    def test_form_raises_error_if_super_secret_without_secret(self):
        post_data = {'name': 'project', 'slug': 'project',
                     'super_secret': True, 'secret': '',
                     'owner': self.user_profile}
        form = ProjectCreateForm(post_data)
        self.assertFalse(form.is_valid())



from django.http import Http404, HttpResponseBadRequest
from django.views.decorators.debug import sensitive_variables

from ddm.models.data_donations import DataDonation
from ddm.models.projects import DonationProject, QuestionnaireResponse, ResearchProfile
from ddm.models.serializers import DonationSerializer, ResponseSerializer, ProjectSerializer

from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework import authentication, permissions


def user_is_allowed_to_download(user, project):
    """
    Returns true if user is owner or collaborator of a project. False otherwise.
    """
    user_profile = ResearchProfile.objects.filter(user=user).first()
    if not user_profile:
        return False
    else:
        if project.owner != user_profile and user_profile not in project.collaborators.all():
            return False
        else:
            return True


class DownloadProjectDataView(APIView):
    """
    View to download project related data donations and questionnaire responses.

    * Token authentication for remote calls.
    * Session authentication for browser access.
    * Only project owners and project collaborators are able to access this view.
    """
    authentication_classes = [authentication.TokenAuthentication, authentication.SessionAuthentication]
    permission_classes = [permissions.IsAuthenticated]

    @sensitive_variables('secret')
    def get(self, request, format=None, *args, **kwargs):
        """
        Return a dictionary object containing the data donations and
        questionnaire responses.
        """
        project_id = self.kwargs['pk']
        project = DonationProject.objects.get(pk=project_id)
        if not user_is_allowed_to_download(request.user, project):
            raise Http404()

        if project.super_secret:
            secret = None if 'super_secret' not in request.headers else request.headers['super_secret']

            if not secret:
                # TODO: Add this to admin log.
                return HttpResponseBadRequest('super_secret required but not found in headers.')

            project.secret_key = secret

        data_donations = DataDonation.objects.filter(project=project)
        q_responses = QuestionnaireResponse.objects.filter(project=project)

        results = {
            'project': ProjectSerializer(project).data,
            'donations': [DonationSerializer(d).data for d in data_donations],
            'responses': [ResponseSerializer(r).data for r in q_responses],
        }
        return Response(results)



from rest_framework import serializers

from ddm.models.data_donations import DataDonation
from ddm.models.projects import DonationProject, QuestionnaireResponse


class ProjectSerializer(serializers.HyperlinkedModelSerializer):

    class Meta:
        model = DonationProject
        fields = ['pk', 'name', 'date_created']


class DonationSerializer(serializers.HyperlinkedModelSerializer):
    project = serializers.IntegerField(source='project.id')
    data = serializers.CharField(source='get_decrypted_data')
    participant = serializers.IntegerField(source='participant.id')

    class Meta:
        model = DataDonation
        fields = ['time_submitted', 'consent', 'status', 'data', 'project', 'participant']


class ResponseSerializer(serializers.HyperlinkedModelSerializer):
    project = serializers.IntegerField(source='project.id')
    data = serializers.CharField(source='get_decrypted_data')
    participant = serializers.IntegerField(source='participant.id')

    class Meta:
        model = QuestionnaireResponse
        fields = ['time_submitted', 'data', 'project', 'participant']



from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from django.http import Http404
from django.urls import reverse
from django.views.generic.edit import CreateView, UpdateView, DeleteView
from django.views.generic.list import ListView

from ddm.forms import BlueprintEditForm
from ddm.models import DonationBlueprint, ZippedBlueprint, DonationInstruction, DonationProject
from ddm.views.project_admin import DdmAuthMixin


class BlueprintMixin:
    """ Mixin for all blueprint related views. """
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({'project_pk': self.kwargs['project_pk']})
        context.update({'project': DonationProject.objects.get(pk=self.kwargs['project_pk'])})
        return context

    def get_success_url(self):
        return reverse('blueprint-list', kwargs={'project_pk': self.kwargs['project_pk']})


class ProjectBlueprintList(DdmAuthMixin, BlueprintMixin, ListView):
    """ View to list all donation blueprints associated with a project. """
    model = DonationBlueprint
    context_object_name = 'donation_blueprints'
    template_name = 'ddm/project_admin/blueprint/list.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({
            'zipped_blueprints': ZippedBlueprint.objects.filter(project_id=self.kwargs['project_pk']),
        })
        return context

    def get_queryset(self):
        queryset = super().get_queryset().filter(project_id=self.kwargs['project_pk'])
        return queryset


class BlueprintCreate(SuccessMessageMixin, DdmAuthMixin, BlueprintMixin, CreateView):
    """ View to create a new donation blueprint. """
    model = DonationBlueprint
    template_name = 'ddm/project_admin/blueprint/create.html'
    fields = ['name', 'exp_file_format', 'expected_fields', 'extracted_fields']
    success_message = 'Blueprint was created successfully.'

    def form_valid(self, form):
        form.instance.project_id = self.kwargs['project_pk']
        return super().form_valid(form)


class BlueprintEdit(SuccessMessageMixin, DdmAuthMixin, BlueprintMixin, UpdateView):
    """ View to edit the details of an existing donation blueprint. """
    model = DonationBlueprint
    template_name = 'ddm/project_admin/blueprint/edit.html'
    form_class = BlueprintEditForm
    success_message = 'Blueprint "%(name)s" was successfully updated.'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        zbp_queryset = ZippedBlueprint.objects.filter(project_id=self.kwargs['project_pk'])
        context['form'].fields['zip_blueprint'].queryset = zbp_queryset
        return context


class BlueprintDelete(DdmAuthMixin, BlueprintMixin, DeleteView):
    """ View to delete an existing donation blueprint. """
    model = DonationBlueprint
    template_name = 'ddm/project_admin/blueprint/delete.html'
    success_message = 'Blueprint "%s" was deleted.'

    def delete(self, request, *args, **kwargs):
        messages.success(self.request, self.success_message % self.get_object().name)
        return super().delete(request, *args, **kwargs)


class ZippedBlueprintCreate(SuccessMessageMixin, DdmAuthMixin, BlueprintMixin, CreateView):
    """ View to create a new zipped blueprint. """
    model = ZippedBlueprint
    template_name = 'ddm/project_admin/blueprint/create.html'
    fields = ['name']
    success_message = 'Zip Blueprint was created successfully.'

    def form_valid(self, form):
        form.instance.project_id = self.kwargs['project_pk']
        return super().form_valid(form)


class ZippedBlueprintEdit(SuccessMessageMixin, DdmAuthMixin, BlueprintMixin, UpdateView):
    """ View to edit the details of an existing zipped blueprint. """
    model = ZippedBlueprint
    template_name = 'ddm/project_admin/blueprint/edit.html'
    fields = ['name']
    success_message = 'Zip blueprint "%(name)s" was successfully updated.'


class ZippedBlueprintDelete(DdmAuthMixin, BlueprintMixin, DeleteView):
    """ View to delete an existing zipped blueprint. """
    model = ZippedBlueprint
    template_name = 'ddm/project_admin/blueprint/delete.html'
    success_message = 'Zip blueprint "%s" was deleted.'

    def delete(self, request, *args, **kwargs):
        messages.success(self.request, self.success_message % self.get_object().name)
        return super().delete(request, *args, **kwargs)


class InstructionMixin:
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({'project_pk': self.kwargs['project_pk']})
        context.update({'project': DonationProject.objects.get(pk=self.kwargs['project_pk'])})
        if self.kwargs['blueprint_type'] == 'blueprint':
            blueprint = DonationBlueprint.objects.get(id=self.kwargs['blueprint_pk'])
            blueprint_type = 'blueprint'
        elif self.kwargs['blueprint_type'] == 'zip-blueprint':
            blueprint = ZippedBlueprint.objects.get(id=self.kwargs['blueprint_pk'])
            blueprint_type = 'zipped Blueprint'
        else:
            raise Http404
        context.update({
            'blueprint': blueprint,
            'blueprint_type': blueprint_type,
        })
        return context

    def get_success_url(self):
        kwargs = {'project_pk': self.kwargs['project_pk'],
                  'blueprint_type': self.kwargs['blueprint_type'],
                  'blueprint_pk': self.kwargs['blueprint_pk']}
        return reverse('instruction-overview', kwargs=kwargs)


class InstructionOverview(DdmAuthMixin, InstructionMixin, ListView):
    """ View to create a new instruction page. """
    model = DonationInstruction
    context_object_name = 'instructions'
    template_name = 'ddm/project_admin/instructions/list.html'
    fields = ['text', 'index']

    def get_queryset(self):
        queryset = super().get_queryset()
        if self.kwargs['blueprint_type'] == 'blueprint':
            queryset = queryset.filter(blueprint_id=self.kwargs['blueprint_pk'])
        elif self.kwargs['blueprint_type'] == 'zip-blueprint':
            queryset = queryset.filter(zip_blueprint_id=self.kwargs['blueprint_pk'])
        else:
            raise Http404
        return queryset


class InstructionCreate(SuccessMessageMixin, DdmAuthMixin, InstructionMixin, CreateView):
    """ View to create an instruction page. """
    model = DonationInstruction
    template_name = 'ddm/project_admin/instructions/create.html'
    fields = ['text', 'index']
    success_message = 'Instruction page was successfully created.'

    def get_blueprint(self):
        if self.kwargs['blueprint_type'] == 'blueprint':
            blueprint = DonationBlueprint.objects.get(id=self.kwargs['blueprint_pk'])
        elif self.kwargs['blueprint_type'] == 'zip-blueprint':
            blueprint = ZippedBlueprint.objects.get(id=self.kwargs['blueprint_pk'])
        return blueprint

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        blueprint = self.get_blueprint()
        if isinstance(blueprint, DonationBlueprint):
            kwargs['instance'] = DonationInstruction(blueprint=blueprint)
        elif isinstance(blueprint, ZippedBlueprint):
            kwargs['instance'] = DonationInstruction(zip_blueprint=blueprint)
        return kwargs

    def get_initial(self):
        initial = super().get_initial()
        indices = self.get_blueprint().donationinstruction_set.values_list('index', flat=True)
        if indices:
            initial['index'] = max(indices) + 1
        else:
            initial['index'] = 1
        return initial


class InstructionEdit(SuccessMessageMixin, DdmAuthMixin, InstructionMixin, UpdateView):
    """ View to edit an instruction page. """
    model = DonationInstruction
    template_name = 'ddm/project_admin/instructions/edit.html'
    fields = ['text', 'index']
    success_message = 'Instruction page was successfully updated.'


class InstructionDelete(DdmAuthMixin, InstructionMixin, DeleteView):
    """ View to delete an instruction page. """
    model = DonationInstruction
    template_name = 'ddm/project_admin/instructions/delete.html'
    success_message = 'Instruction page was deleted.'

    def delete(self, request, *args, **kwargs):
        messages.success(self.request, self.success_message)
        return super().delete(request, *args, **kwargs)

    def get_success_url(self):
        kwargs = {'project_pk': self.kwargs['project_pk'],
                  'blueprint_type': self.kwargs['blueprint_type'],
                  'blueprint_pk': self.kwargs['blueprint_pk']}
        return reverse('instruction-overview', kwargs=kwargs)



from Crypto.Cipher import PKCS1_OAEP, AES
from Crypto.Hash import HMAC
from Crypto.PublicKey import RSA
from Crypto.Random import get_random_bytes
from django.db import models
from io import BytesIO
from struct import pack

import base64
import json


class ModelWithEncryptedData(models.Model):

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        print()
        self.data = Encryption(
            secret=self.project.secret_key,
            salt=str(self.project.date_created),
            public_key=self.project.public_key
        ).encrypt(self.data)
        super().save(*args, **kwargs)

    def get_decrypted_data(self):
        return Encryption(secret=self.project.secret_key, salt=str(self.project.date_created)).decrypt(self.data)


class Encryption:

    class PRNG(object):
        def __init__(self, seed):
            self.index = 0
            self.seed = seed
            self.buffer = b""

        def __call__(self, n):
            while len(self.buffer) < n:
                self.buffer += HMAC.new(self.seed + pack("<I", self.index)).digest()
                self.index += 1
            result, self.buffer = self.buffer[:n], self.buffer[n:]
            return result

    def __init__(self, secret=None, salt=None, public_key=None):
        self.counter = 0
        if public_key:
            self.public_key = public_key
        else:
            seed_128 = HMAC.new(bytes(secret, 'utf-8') + bytes(salt, 'utf-8')).digest()
            self.rsa = RSA.generate(2048, randfunc=Encryption.PRNG(seed_128))
            self.public_key = self.rsa.publickey().exportKey()

    def encrypt(self, value):
        """
        Save encrypted data to the database, either use a supplied 'public_key'
        or generate the material dynamically.
        """
        session_key = get_random_bytes(16)
        cipher_aes = AES.new(session_key, AES.MODE_EAX)
        cipher_data, tag = cipher_aes.encrypt_and_digest(bytes(json.dumps(value), 'utf-8'))
        cipher_rsa = PKCS1_OAEP.new(RSA.importKey(self.public_key))
        encrypted_session_key = cipher_rsa.encrypt(session_key)
        payload = encrypted_session_key + cipher_aes.nonce + tag + cipher_data
        return base64.encodebytes(payload)

    def decrypt(self, value):
        """
        Fetch encrypted data from the database, try to decode the stored data,
        None if decryption fails.
        """
        value_bytes = BytesIO(base64.decodebytes(value))
        encrypted_session_key = value_bytes.read(self.rsa.size_in_bytes())
        nonce = value_bytes.read(16)
        tag = value_bytes.read(16)
        cipher_data = value_bytes.read()
        session_key = PKCS1_OAEP.new(self.rsa).decrypt(encrypted_session_key)
        cipher_aes = AES.new(session_key, AES.MODE_EAX, nonce)
        decrypted = cipher_aes.decrypt_and_verify(cipher_data, tag)
        return json.loads(decrypted.decode('utf-8'))



from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from django.urls import reverse_lazy
from django.views.generic.detail import DetailView
from django.views.generic.edit import CreateView, UpdateView, DeleteView
from django.views.generic.list import ListView

from ddm.forms import ProjectCreateForm
from ddm.models import DonationProject, ResearchProfile, ExceptionLogEntry
from ddm.views.project_admin.auth import DdmAuthMixin


class ProjectList(DdmAuthMixin, ListView):
    """ View to display a list of existing donation projects. """
    model = DonationProject
    template_name = 'ddm/project_admin/project/list.html'

    def get_queryset(self):
        return DonationProject.objects.filter(owner__user=self.request.user)


class ProjectCreate(SuccessMessageMixin, DdmAuthMixin, CreateView):
    """ View to create a new donation project. """
    model = DonationProject
    template_name = 'ddm/project_admin/project/create.html'
    form_class = ProjectCreateForm
    success_message = 'Project was created successfully.'

    def get_initial(self):
        self.initial = super().get_initial()
        self.initial.update({'owner': ResearchProfile.objects.get(user=self.request.user)})
        return self.initial

    def form_valid(self, form):
        form.instance.owner = ResearchProfile.objects.get(user=self.request.user)
        return super().form_valid(form)


class ProjectDetail(DdmAuthMixin, DetailView):
    """ View to display landing page for project. """
    model = DonationProject
    template_name = 'ddm/project_admin/project/detail.html'


class ProjectEdit(SuccessMessageMixin, DdmAuthMixin, UpdateView):
    """ View to edit the details of an existing donation project. """
    model = DonationProject
    template_name = 'ddm/project_admin/project/edit.html'
    fields = ['name', 'slug']
    success_message = 'Project details successfully updated.'


class ProjectDelete(DdmAuthMixin, DeleteView):
    """ View to display a list of existing donation projects. """
    model = DonationProject
    template_name = 'ddm/project_admin/project/delete.html'
    success_url = reverse_lazy('project-list')
    success_message = 'Project was deleted.'

    def delete(self, request, *args, **kwargs):
        messages.success(self.request, self.success_message)
        return super().delete(request, *args, **kwargs)


class WelcomePageEdit(SuccessMessageMixin, DdmAuthMixin, UpdateView):
    """ View to edit the welcome page. """
    model = DonationProject
    template_name = 'ddm/project_admin/project/edit-welcome.html'
    fields = ['intro_text']
    success_message = 'Welcome Page successfully updated.'


class EndPageEdit(SuccessMessageMixin, DdmAuthMixin, UpdateView):
    """ View to edit the welcome page. """
    model = DonationProject
    template_name = 'ddm/project_admin/project/edit-end.html'
    fields = ['outro_text']
    success_message = 'End Page successfully updated.'


class ExceptionList(SuccessMessageMixin, DdmAuthMixin, ListView):
    """ View that lists all exceptions related to a project. """
    model = ExceptionLogEntry
    template_name = 'ddm/project_admin/project/exceptions.html'
    context_object_name = 'exceptions'

    def get_queryset(self):
        qs = super().get_queryset()
        project_id = self.kwargs.get('project_pk')
        if project_id is not None:
            project = DonationProject.objects.get(pk=project_id)
            qs = qs.filter(project=project).order_by('-date')
        return qs

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        project_id = self.kwargs.get('project_pk')
        if project_id is not None:
            project = DonationProject.objects.get(pk=project_id)
        else:
            project = None
        context.update({'project': project})
        return context



from django.db import models
from ddm.models import DonationProject, Participant


class ExceptionLogEntry(models.Model):
    date = models.DateTimeField()
    project = models.ForeignKey(DonationProject, on_delete=models.CASCADE)
    participant = models.ForeignKey(Participant, null=True, on_delete=models.SET_NULL)
    exception_type = models.IntegerField()
    message = models.TextField()



from django.urls import reverse
from ddm.models import DonationProject
from ddm.tests.base import TestData


class BaseAdminViewTestMixin:
    def test_logged_out_redirects_to_login_view(self):
        response = self.client.get(self.url, follow=True)
        self.assertRedirects(response, reverse('ddm-login'))

    def test_superuser_logged_in_returns_200(self):
        self.client.login(**self.users['super']['credentials'])
        response = self.client.get(self.url, follow=True)
        self.assertEqual(response.status_code, 200)

    def test_owner_logged_in_returns_200(self):
        self.client.login(**self.users['base']['credentials'])
        response = self.client.get(self.url, follow=True)
        self.assertEqual(response.status_code, 200)

    def test_non_owner_logged_in_returns_404(self):
        if self.url in [reverse('project-list'), reverse('project-create')]:
            pass
        else:
            self.client.login(**self.users['base2']['credentials'])
            response = self.client.get(self.url, follow=True)
            self.assertEqual(response.status_code, 404)

    def test_user_without_profile_logged_in_redirects_to_registration_view(self):
        self.client.login(**self.users['no_profile']['credentials'])
        response = self.client.get(self.url, follow=True)
        self.assertRedirects(response, reverse('ddm-register'))

    def test_user_without_permission_logged_in_redirects_to_no_permission_view(self):
        if self.url in [reverse('project-list'), reverse('project-create')]:
            pass
        else:
            self.client.login(**self.users['no_permission']['credentials'])
            response = self.client.get(self.url_no_perm, follow=True)
            self.assertRedirects(response, reverse('ddm-no-permission'))


class ProjectListViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('project-list', args=[])

    def test_correct_queryset_returned(self):
        self.client.login(**self.users['base']['credentials'])
        response = self.client.get(reverse('project-list'))
        object_list = response.context['object_list']
        expected_queryset = DonationProject.objects.filter(owner__user=self.users['base']['user'])
        self.assertQuerysetEqual(object_list, list(expected_queryset), ordered=False)


class ProjectCreateViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('project-create', args=[])

    # TODO: Check this test again: Something is not working as expected with the client.post() call.
    #  The View works as expected when called in browser (runserver).
    # def test_project_is_created(self):
    #     self.client.login(**self.users['base']['credentials'])
    #     self.client.post(reverse('project-create'), data={'name': 'name', 'slug': 'new_slug'})
    #     self.assertTrue(DonationProject.objects.filter(
    #         owner=self.users['base']['profile'], slug='new_slug').exists())


class ProjectEditViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('project-edit', args=[self.project_base.pk])
        self.url_no_perm = reverse('project-edit', args=[self.project_no_perm.pk])


class ProjectDeleteViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('project-delete', args=[self.project_base.pk])
        self.url_no_perm = reverse('project-delete', args=[self.project_no_perm.pk])


class ProjectWelcomeEditViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('welcome-page-edit', args=[self.project_base.pk])
        self.url_no_perm = reverse('welcome-page-edit', args=[self.project_no_perm.pk])


class ProjectEndEditViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('end-page-edit', args=[self.project_base.pk])
        self.url_no_perm = reverse('end-page-edit', args=[self.project_no_perm.pk])


class BlueprintListViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('blueprint-list', args=[self.project_base.pk])
        self.url_no_perm = reverse('blueprint-list', args=[self.project_no_perm.pk])


class BlueprintCreateViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('blueprint-create', args=[self.project_base.pk])
        self.url_no_perm = reverse('blueprint-create', args=[self.project_no_perm.pk])


class BlueprintEditViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('blueprint-edit', args=[self.project_base.pk, self.don_bp.pk])
        self.url_no_perm = reverse('blueprint-edit', args=[self.project_no_perm.pk, self.don_bp.pk])


class BlueprintDeleteViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('blueprint-delete', args=[self.project_base.pk, self.don_bp.pk])
        self.url_no_perm = reverse('blueprint-delete', args=[self.project_no_perm.pk, self.don_bp.pk])


class ZippedBlueprintCreateViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('zipped-blueprint-create', args=[self.project_base.pk])
        self.url_no_perm = reverse('zipped-blueprint-create', args=[self.project_no_perm.pk])


class ZippedBlueprintEditViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('zipped-blueprint-edit', args=[self.project_base.pk, self.zip_bp.pk])
        self.url_no_perm = reverse('zipped-blueprint-edit', args=[self.project_no_perm.pk, self.zip_bp.pk])


class ZippedBlueprintDeleteViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('zipped-blueprint-delete', args=[self.project_base.pk, self.zip_bp.pk])
        self.url_no_perm = reverse('zipped-blueprint-delete', args=[self.project_no_perm.pk, self.zip_bp.pk])


class InstructionListViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('instruction-overview', args=[self.project_base.pk, 'blueprint', self.don_bp.pk])
        self.url_no_perm = reverse('instruction-overview', args=[self.project_no_perm.pk, 'blueprint', self.don_bp.pk])


class InstructionCreateViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('instruction-create', args=[self.project_base.pk, 'blueprint', self.don_bp.pk])
        self.url_no_perm = reverse('instruction-create', args=[self.project_no_perm.pk, 'blueprint', self.don_bp.pk])


class InstructionEditViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('instruction-edit',
                           args=[self.project_base.pk, 'blueprint', self.don_bp.pk, self.instruction.pk])
        self.url_no_perm = reverse('instruction-edit',
                                   args=[self.project_no_perm.pk, 'blueprint', self.don_bp.pk, self.instruction.pk])


class InstructionDeleteViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('instruction-delete',
                           args=[self.project_base.pk, 'blueprint', self.don_bp.pk, self.instruction.pk])
        self.url_no_perm = reverse('instruction-delete',
                                   args=[self.project_no_perm.pk, 'blueprint', self.don_bp.pk, self.instruction.pk])


class QuestionnaireOverviewViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('questionnaire-overview',  args=[self.project_base.pk])
        self.url_no_perm = reverse('questionnaire-overview', args=[self.project_no_perm.pk])


class QuestionCreateViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.question_type_slugs = [
            'single_choice',
            'multi_choice',
            'open',
            'matrix',
            'semantic_diff',
            'transition'
        ]

    def test_logged_out_redirects_to_login_view(self):
        for slug in self.question_type_slugs:
            url = reverse('question-create', args=[self.project_base.pk, slug])
            response = self.client.get(url, follow=True)
            self.assertRedirects(response, reverse('ddm-login'))

    def test_superuser_logged_in_returns_200(self):
        self.client.login(**self.users['super']['credentials'])
        for slug in self.question_type_slugs:
            url = reverse('question-create', args=[self.project_base.pk, slug])
            response = self.client.get(url, follow=True)
            self.assertEqual(response.status_code, 200)

    def test_owner_logged_in_returns_200(self):
        self.client.login(**self.users['base']['credentials'])
        for slug in self.question_type_slugs:
            url = reverse('question-create', args=[self.project_base.pk, slug])
            response = self.client.get(url, follow=True)
            self.assertEqual(response.status_code, 200)

    def test_non_owner_logged_in_returns_404(self):
        self.client.login(**self.users['base2']['credentials'])
        for slug in self.question_type_slugs:
            url = reverse('question-create', args=[self.project_base.pk, slug])
            response = self.client.get(url, follow=True)
            self.assertEqual(response.status_code, 404)

    def test_user_without_profile_logged_in_redirects_to_registration_view(self):
        self.client.login(**self.users['no_profile']['credentials'])
        for slug in self.question_type_slugs:
            url = reverse('question-create', args=[self.project_base.pk, slug])
            response = self.client.get(url, follow=True)
            self.assertRedirects(response, reverse('ddm-register'))

    def test_user_without_permission_logged_in_redirects_to_no_permission_view(self):
        self.client.login(**self.users['no_permission']['credentials'])
        for slug in self.question_type_slugs:
            url = reverse('question-create', args=[self.project_no_perm.pk, slug])
            response = self.client.get(url, follow=True)
            self.assertRedirects(response, reverse('ddm-no-permission'))


class QuestionEditViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('question-edit', args=[self.project_base.pk, 'open', self.open_quest.pk])
        self.url_no_perm = reverse('question-edit', args=[self.project_no_perm.pk, 'open', self.open_quest.pk])


class QuestionDeleteViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('question-delete', args=[self.project_base.pk, 'open', self.open_quest.pk])
        self.url_no_perm = reverse('question-delete', args=[self.project_no_perm.pk, 'open', self.open_quest.pk])


class QuestionItemsEditViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('question-items', args=[self.project_base.pk, 'single_choice', self.sc_quest.pk])
        self.url_no_perm = reverse('question-items',
                                   args=[self.project_no_perm.pk, 'single_choice', self.sc_quest.pk])


class QuestionScaleEditViewTests(BaseAdminViewTestMixin, TestData):
    def setUp(self):
        super().setUp()
        self.url = reverse('question-scale', args=[self.project_base.pk, 'matrix', self.matrix_quest.pk])
        self.url_no_perm = reverse('question-scale', args=[self.project_no_perm.pk, 'matrix', self.matrix_quest.pk])



from django.urls import reverse
from django.utils import timezone

from rest_framework.test import APIClient

from ddm.models import DataDonation, QuestionnaireResponse
from ddm.tests.base import TestData


class TestDownload(TestData):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        DataDonation.objects.create(
            project=cls.project_base,
            blueprint=cls.don_bp,
            participant=cls.participant_base,
            time_submitted=timezone.now(),
            consent=True,
            status='{}',
            data='{"data": ["donated_data", "donated_data"]}'
        )
        QuestionnaireResponse.objects.create(
            project=cls.project_base,
            participant=cls.participant_base,
            time_submitted=timezone.now(),
            data='{"data": ["response_data", "response_data"]}'
        )
        cls.project_base.collaborators.add(cls.users['base2']['profile'])
        cls.project_base.save()

    def test_download_project_data_view_exists(self):
        response = self.client.get(reverse('ddm-download-api', args=[self.project_base.pk]))
        self.assertEqual(response.status_code, 401)

    def test_download_with_regular_login_owner(self):
        self.client.login(**self.users['base']['credentials'])
        response = self.client.get(reverse('ddm-download-api', args=[self.project_base.pk]))
        self.assertEqual(response.status_code, 200)
        self.assertIsNotNone(response.data)

    def test_download_with_regular_login_collaborator(self):
        self.client.login(**self.users['base2']['credentials'])
        response = self.client.get(reverse('ddm-download-api', args=[self.project_base.pk]))
        self.assertEqual(response.status_code, 200)
        self.assertIsNotNone(response.data)

    def test_download_fails_for_user_without_permission(self):
        self.client.login(**self.users['base3']['credentials'])
        response = self.client.get(reverse('ddm-download-api', args=[self.project_base.pk]), follow=True)
        self.assertEqual(response.status_code, 404)

    def test_download_with_valid_api_credentials(self):
        token = self.users['base']['profile'].create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        response = client.get(reverse('ddm-download-api', args=[self.project_base.pk]))
        self.assertEqual(response.status_code, 200)
        self.assertIsNotNone(response.data)

    def test_download_with_invalid_api_credentials(self):
        token = 'adsjfjklsdjflkjasflk'
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token)
        response = client.get(reverse('ddm-download-api', args=[self.project_base.pk]))
        self.assertEqual(response.status_code, 401)



from ckeditor.fields import RichTextField

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.urls import reverse
from django.utils import timezone

from ddm.models import Encryption, ModelWithEncryptedData

from rest_framework.authtoken.models import Token


class ResearchProfile(models.Model):
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name='ddm_research_profile',
    )
    active = models.BooleanField(default=True)
    created = models.DateTimeField('date registered', default=timezone.now)
    ignore_email_restriction = models.BooleanField(default=False)

    def get_token(self):
        return Token.objects.filter(user=self.user).first()

    def create_token(self):
        token = self.get_token()
        if token:
            token.delete()
        return Token.objects.create(user=self.user)


class DonationProject(models.Model):
    name = models.CharField(max_length=50)
    slug = models.SlugField(unique=True, verbose_name='External Project Slug')
    intro_text = RichTextField(null=True, blank=True, verbose_name='Welcome Page Text')
    outro_text = RichTextField(null=True, blank=True, verbose_name='End Page Text')

    date_created = models.DateTimeField(default=timezone.now)

    public_key = models.BinaryField()
    super_secret = models.BooleanField(default=False)

    owner = models.ForeignKey(
        ResearchProfile,
        related_name='project_owner',
        on_delete=models.SET_NULL,
        null=True
    )
    collaborators = models.ManyToManyField(
        ResearchProfile,
        related_name='project_collaborators'
    )

    def __init__(self, *args, **kwargs):
        self.secret_key = settings.SECRET_KEY
        if 'secret_key' in kwargs:
            self.secret_key = kwargs['secret_key']
            kwargs.pop('secret_key')
        super().__init__(*args, **kwargs)

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse('project-detail', args=[str(self.id)])

    def save(self, *args, **kwargs):
        if not self.pk:
            if self.owner is None:
                raise ValidationError(
                    'DonationProject.owner cannot be null on create.')

            if self.super_secret:
                self.public_key = Encryption(self.secret, str(self.date_created)).public_key
        else:
            if self.owner in self.collaborators.all():
                raise ValidationError(
                    'DonationProject.owner cannot be a collaborator at the same time.')
        super().save(*args, **kwargs)

    @property
    def secret(self):
        return self.secret_key

    @secret.setter
    def secret(self, value):
        self.secret_key = value


class Participant(models.Model):
    project = models.ForeignKey(DonationProject, on_delete=models.CASCADE)

    # Participation statistics.
    start_time = models.DateTimeField()
    end_time = models.DateTimeField(null=True)
    completed = models.BooleanField(default=False)


class QuestionnaireResponse(ModelWithEncryptedData):
    # Will only ever be deleted, when the project is deleted.
    project = models.ForeignKey(DonationProject, on_delete=models.CASCADE)
    participant = models.ForeignKey(Participant, on_delete=models.CASCADE)
    time_submitted = models.DateTimeField(default=timezone.now)
    data = models.BinaryField()



from .encryption import *
from .projects import (
    DonationProject, Participant, QuestionnaireResponse, ResearchProfile
)
from .data_donations import (
    DonationBlueprint, ZippedBlueprint, DataDonation, DonationInstruction
)
from .questions import *
from .exceptions import *



import json

from ckeditor.fields import RichTextField

from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator, MinValueValidator
from django.db import models
from django.urls import reverse
from django.utils import timezone

from ddm.models import DonationProject, Participant, ModelWithEncryptedData


import logging
logger = logging.getLogger(__name__)

COMMA_SEPARATED_STRINGS_VALIDATOR = RegexValidator(
    r'^((["][^"]+["]))(\s*,\s*((["][^"]+["])))*[,\s]*$',
    message=(
        'Field must contain one or multiple comma separated strings. '
        'Strings must be enclosed in double quotes ("string").'
    )
)


class ZippedBlueprint(models.Model):
    name = models.CharField(max_length=250)
    project = models.ForeignKey(
        DonationProject,
        on_delete=models.CASCADE
    )

    def __str__(self):
        return self.name

    def get_slug(self):
        return 'zip-blueprint'

    def get_absolute_url(self):
        return reverse('zipped-blueprint-edit', args=[str(self.project_id), str(self.id)])

    def get_blueprints(self):
        blueprints = DonationBlueprint.objects.filter(zip_blueprint=self)
        return blueprints

    def get_configs(self):
        bp_configs = []
        blueprints = self.get_blueprints()
        for bp in blueprints:
            bp_configs.append(bp.get_config())
        return bp_configs

    def get_instructions(self):
        return [{'index': i.index, 'text': i.text} for i in self.donationinstruction_set.all()]


# TODO: For admin section: Add validation on save to ensure that regex_path != None when zip_blueprint != None
class DonationBlueprint(models.Model):
    project = models.ForeignKey(
        DonationProject,
        on_delete=models.CASCADE
    )

    name = models.CharField(max_length=250)

    class FileFormats(models.TextChoices):
        JSON_FORMAT = 'json'
        # CSV_FORMAT = 'csv',
        # HTML_FORMAT = 'html',
        # XLSX_FORMAT = 'xlsx',

    exp_file_format = models.CharField(
        max_length=10,
        choices=FileFormats.choices,
        default=FileFormats.JSON_FORMAT,
        verbose_name='Expected file format',
    )

    expected_fields = models.TextField(
        null=False,
        blank=False,
        validators=[COMMA_SEPARATED_STRINGS_VALIDATOR],
        help_text='Put the field names in double quotes (") and separate them with commas ("Field A", "Field B").'
    )
    extracted_fields = models.TextField(
        null=True,
        blank=True,
        validators=[COMMA_SEPARATED_STRINGS_VALIDATOR],
        help_text='Put the field names in double quotes (") and separate them with commas ("Field A", "Field B").'
    )

    # Configuration if related to ZippedBlueprint:
    zip_blueprint = models.ForeignKey(
        ZippedBlueprint,
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name='Zip container',
    )
    regex_path = models.TextField(null=True, blank=True)

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse('blueprint-edit', args=[str(self.project_id), str(self.id)])

    def get_slug(self):
        return 'blueprint'

    def get_config(self):
        config = {
            'id': self.pk,
            'name': self.name,
            'format': self.exp_file_format,
            'f_expected': json.loads("[" + str(self.expected_fields) + "]"),
            'f_extract': json.loads("[" + str(self.extracted_fields) + "]"),
            'regex_path': self.regex_path,
        }
        return config

    def get_instructions(self):
        return [{'index': i.index, 'text': i.text} for i in self.donationinstruction_set.all()]

    def get_associated_questions(self):
        return self.questionbase_set.all()

    def process_donation(self, data, participant):
        if self.validate_donation(data):
            self.create_donation(data, participant)
        else:
            logger.error(f'Donation not processed by blueprint {self.pk}')
        return

    @staticmethod
    def validate_donation(data):
        # Check if all expected fields are in response.
        response_fields = ['consent', 'extracted_data', 'status']
        if not all(k in data for k in response_fields):
            logger.error(
                'Donation data does not contain the expected information. '
                f'Expected key: {response_fields}; '
                f'Present fields: {data.keys()}.'
            )
            return False

        # TODO: Add other validation steps? - e.g., validation of extracted fields
        return True

    def create_donation(self, data, participant):
        DataDonation.objects.create(
            project=self.project,
            blueprint=self,
            participant=participant,
            consent=data['consent'],
            status=data['status'],
            data=data['extracted_data']
        )
        return


class DataDonation(ModelWithEncryptedData):
    project = models.ForeignKey(
        DonationProject,
        on_delete=models.CASCADE
    )
    blueprint = models.ForeignKey(
        DonationBlueprint,
        null=True,
        on_delete=models.SET_NULL
    )
    participant = models.ForeignKey(
        Participant,
        on_delete=models.CASCADE
    )
    time_submitted = models.DateTimeField(default=timezone.now)
    consent = models.BooleanField(default=False)
    status = models.JSONField()
    data = models.BinaryField()


class DonationInstruction(models.Model):
    text = RichTextField(null=True, blank=True)
    index = models.PositiveIntegerField(default=1, validators=[MinValueValidator(1)])
    blueprint = models.ForeignKey(
        DonationBlueprint,
        null=True,
        blank=True,
        on_delete=models.CASCADE
    )
    zip_blueprint = models.ForeignKey(
        ZippedBlueprint,
        null=True,
        blank=True,
        on_delete=models.CASCADE
    )

    class Meta:
        ordering = ['index']
        constraints = [
            models.UniqueConstraint(
                fields=['index', 'blueprint'],
                name='unique_index_per_blueprint'
            ),
            models.UniqueConstraint(
                fields=['index', 'zip_blueprint'],
                name='unique_index_per_zipblueprint'
            ),
        ]

    def get_query_object(self):
        if self.blueprint:
            query_object = self.blueprint
        else:
            query_object = self.zip_blueprint
        return query_object

    def clean(self):
        # Ensure that instruction is correctly linked to one blueprint type.
        if not self.blueprint and not self.zip_blueprint:
            raise ValidationError(
                'Must be linked to either a DonationBlueprint or '
                'a ZippedBlueprint.'
            )
        if self.blueprint and self.zip_blueprint:
            raise ValidationError(
                'Must be linked to either a DonationBlueprint or '
                'a ZippedBlueprint, but not both.'
            )

        # Ensure that index of instruction page is not greater than set of existing instructions + 1.
        n_instructions = self.get_query_object().donationinstruction_set.count()
        if self.pk:
            if self.index > n_instructions:
                raise ValidationError(
                    f'Index must be in range 1 to {n_instructions}.'
                )
        else:
            if self.index > (n_instructions + 1):
                raise ValidationError(
                    f'Index must be in range 1 to {n_instructions + 1}.'
                )
        super().clean()

    def save(self, *args, **kwargs):
        if kwargs.pop('ignore_index_check', False):
            return super().save()

        # TODO: Optimize and prettify the following part:
        initial_index = DonationInstruction.objects.get(pk=self.pk).index if self.pk else None
        index_taken = self.get_query_object().donationinstruction_set.filter(index=self.index).exclude(pk=self.pk).exists()
        if index_taken and (self.index != initial_index):

            # Account for unique constraint by doing a "proxy"-save to free index.
            target_index = self.index
            self.index = self.get_query_object().donationinstruction_set.count() + 5
            super().save()

            # Change indices of involved objects:
            queryset = self.get_query_object().donationinstruction_set.exclude(pk=self.pk)
            if initial_index is None:
                queryset = queryset.filter(index__gte=target_index).order_by('-index')
                for q in queryset:
                    q.index += 1
                for q in queryset:
                    q.save(ignore_index_check=True)
            elif target_index < initial_index:
                queryset = queryset.filter(index__gte=target_index, index__lt=initial_index).order_by('-index')
                for q in queryset:
                    q.index += 1
                for q in queryset:
                    q.save(ignore_index_check=True)
            elif target_index > initial_index:
                queryset = queryset.filter(index__gt=initial_index, index__lte=target_index).order_by('index')
                for q in queryset:
                    q.index -= 1
                for q in queryset:
                    q.save(ignore_index_check=True)

            # Revert "proxy"-save.
            self.index = target_index
            return super().save()

        return super().save()

    def delete(self, *args, **kwargs):
        """ This model has a post_delete signal processor (see signals.py). """
        super().delete(*args, **kwargs)



from django.db.models.signals import post_delete
from django.dispatch import receiver

from ddm.models import DonationInstruction


@receiver(post_delete, sender=DonationInstruction)
def reorder_indices(sender, instance, **kwargs):
    queryset = instance.get_query_object().donationinstruction_set.filter(index__gt=instance.index).order_by('index')
    for q in queryset:
        q.index -= 1
        q.save()
    return



from ddm.models import DonationProject, ExceptionLogEntry, Participant
from django.utils import timezone
from rest_framework import permissions
from rest_framework.response import Response
from rest_framework.views import APIView


class ExceptionAPI(APIView):

    permission_classes = [permissions.AllowAny]

    def post(self, request, format=None, *args, **kwargs):
        """
        Except an error message and log it
        """
        project_id = self.kwargs['pk']
        project = DonationProject.objects.get(pk=project_id)

        participant_id = request.session['projects'][f'{project_id}']['participant_id']
        participant = Participant.objects.get(pk=participant_id)

        ExceptionLogEntry.objects.create(
            date=timezone.now(),
            project=project,
            participant=participant,
            exception_type=request.data['status_code'],
            message=request.data['message']
        )

        return Response(None, status=201)



from django.test import TestCase

from ddm.models import DataDonation, DonationProject, Encryption, QuestionnaireResponse
from ddm.tests.base import TestData


class TestEncryptedFieldIsolated(TestCase):
    def test_encrypt_decrypt(self):
        text = 'test_string'
        enc = Encryption(secret='foo', salt='bar')
        enc_text = enc.encrypt(text)
        self.assertNotEqual(text, enc_text)
        self.assertEqual(text, enc.decrypt(enc_text))

    def test_encrypt_decrypt_explicit_public(self):
        text = 'test_string'
        enc = Encryption(secret='foo', salt='bar')
        pub = enc.public_key
        enc_text = Encryption(public_key=pub).encrypt(text)
        self.assertNotEqual(text, enc_text)
        self.assertEqual(text, enc.decrypt(enc_text))


class TestModelEncryption(TestData):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.custom_project = DonationProject.objects.create(
            name='Test Project Custom',
            slug='test-project-custom',
            super_secret=True,
            secret_key='test1234',
            owner=cls.users['base']['profile']
        )
        cls.raw_data_short = '{"some_data": "some_value"}'
        cls.raw_data_long = '{' + 100*'"some_data": "some_value",' + '}'

    def test_data_donation_encryption_default(self):
        for raw_data in [self.raw_data_short, self.raw_data_long]:
            with self.subTest(raw_data=raw_data):
                dd = DataDonation.objects.create(
                    project=self.project_base,
                    blueprint=self.don_bp,
                    participant=self.participant_base,
                    consent=True,
                    status='some status',
                    data=raw_data
                )
                self.assertNotEqual(raw_data, dd.data)
                self.assertEqual(raw_data, dd.get_decrypted_data())

    def test_questionnaire_response_encryption_default(self):
        for raw_data in [self.raw_data_short, self.raw_data_long]:
            with self.subTest(raw_data=raw_data):
                qr = QuestionnaireResponse.objects.create(
                    project=self.project_base,
                    participant=self.participant_base,
                    data=raw_data
                )
                self.assertNotEqual(raw_data, qr.data)
                self.assertEqual(raw_data, qr.get_decrypted_data())

    def test_data_donation_encryption_custom(self):
        for raw_data in [self.raw_data_short, self.raw_data_long]:
            with self.subTest(raw_data=raw_data):
                dd = DataDonation.objects.create(
                    project=self.custom_project,
                    blueprint=self.don_bp,
                    participant=self.participant_base,
                    consent=True,
                    status='some status',
                    data=raw_data,
                )
                self.assertNotEqual(raw_data, dd.data)
                self.assertEqual(raw_data, dd.get_decrypted_data())

    def test_questionnaire_response_encryption_custom(self):
        for raw_data in [self.raw_data_short, self.raw_data_long]:
            with self.subTest(raw_data=raw_data):
                qr = QuestionnaireResponse.objects.create(
                    project=self.custom_project,
                    participant=self.participant_base,
                    data=raw_data
                )
                self.assertNotEqual(raw_data, qr.data)
                self.assertEqual(raw_data, qr.get_decrypted_data())



from ddm.models import DonationInstruction
from ddm.tests.base import TestData


class TestDonationInstructionModel(TestData):
    def setUp(self):
        self.inst_2 = DonationInstruction.objects.create(
            text='blueprint 2',
            index=2,
            blueprint=self.don_bp
        )
        self.inst_3 = DonationInstruction.objects.create(
            text='blueprint 3',
            index=3,
            blueprint=self.don_bp
        )

    def test_create_with_existing_index_pushes_other_indices(self):
        new_instruction = DonationInstruction.objects.create(
            text='blueprint new',
            index=2,
            blueprint=self.don_bp
        )
        self.assertEqual(new_instruction.index, 2)
        self.assertEqual(DonationInstruction.objects.get(text='blueprint 2', blueprint=self.don_bp).index, 3)
        self.assertEqual(DonationInstruction.objects.get(text='blueprint 3', blueprint=self.don_bp).index, 4)

    def test_decrease_index_of_existing_instruction(self):
        self.inst_3.index = 2
        self.inst_3.save()
        self.assertEqual(self.inst_3.index, 2)
        self.assertEqual(DonationInstruction.objects.get(text='blueprint 2', blueprint=self.don_bp).index, 3)

    def test_increase_index_of_existing_instruction(self):
        self.inst_2.index = 3
        self.inst_2.save()
        self.assertEqual(self.inst_2.index, 3)
        self.assertEqual(DonationInstruction.objects.get(text='blueprint 3', blueprint=self.don_bp).index, 2)

    def test_indices_are_adjusted_on_delete(self):
        self.inst_2.delete()
        self.assertEqual(DonationInstruction.objects.get(text='blueprint 3', blueprint=self.don_bp).index, 2)
